/* (c) by G. Caronni in 1995			19.11.95 */

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <malloc.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "cert_defs.h"
#include "cert_udp.h"
#include "cert_db.h"
#include "cert_int.h"
#include "cert_ubs.h"
#include "md5.h"

#ifdef USE_SUNCERT
#include "suncert/enlink.h"
#endif


int send_socket;
int recv_socket;


int udp_init(void)
{
    struct sockaddr_in addr;

    if ((send_socket = socket(PF_INET,SOCK_DGRAM,0)) < 0) return -1;
    memset(&addr,0,sizeof (addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(SEND_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(send_socket,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
        close(send_socket);
        perror("CERT udp_init bind send_socket");
        return -1;
    }
    if ((recv_socket = socket(PF_INET,SOCK_DGRAM,0)) < 0) {
        close(send_socket);
	return -1;
    }
    memset(&addr,0,sizeof (addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(RECV_PORT);
    addr.sin_addr.s_addr = htonl(INADDR_ANY);
    if (bind(recv_socket,(struct sockaddr *) &addr,sizeof(addr)) < 0) {
        perror("CERT udp_init bind recv_socket");
        close(send_socket);
        close(recv_socket);
        return -1;
    }
    return 0;
}

void udp_exit(void)
{
    close(send_socket);
    close(recv_socket);
}

static dh_data *add_cert(u_char *buf, struct in_addr from)
{
    dh_data *res=NULL;

#ifdef USE_SUNCERT

    void *cdata;

    if (buf[1]!=8 || *buf==4) {
	cdata=enlink_readudp(buf);

	/* fill in RMK as defined by cert and buf[1] */

	/* validate with correct name and times */

	/* fill in all Bignum and date fields */

	/* THEN go ahead  XXX  */

    } else {
#if 1
        printf("CERT add_cert unknown CERT-TYPE or NSID.\n");
        return NULL;
#endif
    }

#else

    u_char *t=buf;
    u_char *hex;
    int ll,l,i;
    struct MD5Context ctxt;


    if (*t++ != 4 || *t++ != 8) {
#if 0
      printf("CERT add_cert unknown CERT-TYPE or NSID.\n");
#endif
        /* Huh? What? No comprende! */
        return NULL;
    }
    ll=*t++<<8;
    ll+=*t++;
    res=db_cert_init();
    res->rns=8; 
    res->begins=*t++<<24;
    res->begins+=*t++<<16;
    res->begins+=*t++<<8;
    res->expires+=*t++;
    res->expires=*t++<<24;
    res->expires+=*t++<<16;
    res->expires+=*t++<<8;
    res->expires+=*t++;
    l=*t++<<8;
    l+=*t++;
    hex=malloc(2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->p) int_init(&(res->p));
    int_set_str(res->p,hex,16);
    l=*t++<<8;
    l+=*t++;
    hex=realloc(hex,2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->g) int_init(&(res->g));
    int_set_str(res->g,hex,16);
    l=*t++<<8;
    l+=*t++;
    hex=realloc(hex,2*l+1);
    for(i=0;i<l;i++,t++) sprintf(hex+2*i,"%02X",*t);
    if (!res->ki) int_init(&(res->ki));
    int_set_str(res->ki,hex,16);
    free(hex);
    if (t-buf-4!=ll) {
#if 0
        printf("CERT add_cert incoherent length.\n");
#endif
        db_cert_drop(res);
        return NULL;
    }
    res->rmk=malloc(16);
/*printf("%s",int_extract_hex(res->ki));*/
#if 0
    very buggy!
    int_extract_raw(&hex,res->ki,(int_bitsize(res->p)+7)/8);
    cert_MD5Init(&ctxt);
    cert_MD5Update(&ctxt,hex,(int_bitsize(res->p)+7)/8);
    cert_MD5Final(res->rmk,&ctxt);
    free(hex);
#endif
    cert_MD5Init(&ctxt);
    cert_MD5Update(&ctxt,buf+4,t-buf-4);
    cert_MD5Final(res->rmk,&ctxt);

    db_unique_disk_save(&res,inet_ntoa(from),0);

#endif /* !USE_SUNCERT */

    return res;

}

void udp_process(int sock)
{
    extern void do_discovered(u_char nsid, u_char *mkid, dh_data *remote, 
    								u_long ip);
    dh_data *response=NULL;
    u_char buf[20000], *t;
    struct sockaddr_in addr;
    int len, addrlen=sizeof(addr);
    int onr,nr,i;

    len=recvfrom(sock, buf, sizeof(buf), 0, (struct sockaddr *) &addr, 
								&addrlen);
#if 0
    printf("CERT recvfrom(%s) ACT=%d STAT=%d NSID=%d NR=%d\n",
           inet_ntoa(addr.sin_addr),*buf&15,buf[1],buf[2],buf[3]);
#endif
    if (len<=0 || len <4) return;

if ((*buf&15)==15) {
    ubs_process(sock, buf, len, &addr);
}

#if 0
    if ((*buf&15)==1 && buf[2]) {
        printf("     MKID=");
        for(i=0;i<mkid_size[*(buf+2)];i++) printf("%02X", buf[4+i]);
        printf("\n");
    }
#endif

    t=buf;
    if ((*t & 0xf0) != 0x10) return; /* Version 1 only... */
    t=buf+3;
    onr=nr=*t++;
    t+=mkid_size[*(buf+2)];
    while(nr--) {
        dh_data *last;
        int l;
        /* response */
        /* if positive response, find/create remote certificate and dh_data,
         * if the incoming data corresponds to an already existing cert,
         * it may be that it just contains new certificate data, which should
         * be appended. if different public value certificate with same name 
         * already exists, store the new one nevertheless, should later check 
         * validity times, and refuse the new one if overlap 
         * (or remove old one)
         * pass it with nsid, mkid, ip to do_discovered
         * if failure, just pass remote=NULL to do_discovered
         */
        last=add_cert(t,addr.sin_addr);
        if (last && last->rns == *(buf+2) && !memcmp(last->rmk,buf+4,
        						mkid_size[last->rns])) 
	    response=last;
        /* the memcmp also took care of cheating nsid8 replies, as saving */
        /* happens under true name anyway */
        t+=2;
	l=*t++*256;
	l+=*t++;
	t+=l;
    }
    t=buf;
    if (onr) {
        /* positive OR negative response for do_discovered */
        do_discovered(*(t+2), t+4, response, addr.sin_addr.s_addr);
    }
#ifdef USE_SUNCERT
    if ((*t & 0x0f) == 0x01) {
        /* request */
        /* if not nsid 8 or 1, return error 3 */

        if (*(t+2)==1) {
            dh_data *find=db_public_find_load(NULL,*(t+2),t+4);
	    if (find == (dh_data *)(-1)) 
		find=db_public_find_load(find,*(t+2),t+4);
            if (find==NULL || find->cert_data==NULL) {
		*t=(1<<4)|2;
		*(t+1)=2;
		*(t+3)=0;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),
							*buf&15,buf[1],buf[3]);
#endif
		sendto(sock, buf, 4+mkid_size[*(t+2)], 0,
            				(struct sockaddr *) &addr, addrlen);
            } else {
		/* go and send type 1 cert */

/*

oh.  the certificate format has changed. nice to know....

slightly late to hack this too! XXX

*/


        } else if (*t(t+2)==8) {
#else
        if (*t(t+2)==8) {
#endif
	    /* find/load requested public value and send it, or an error */
            dh_data *find=db_public_find_load(NULL,*(t+2),t+4);
	    if (find == (dh_data *)(-1)) 
		find=db_public_find_load(find,*(t+2),t+4);
            if (!find) {
		*t=(1<<4)|2;
		*(t+1)=2;
		*(t+3)=0;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),
							*buf&15,buf[1],buf[3]);
#endif
		sendto(sock, buf, 4+mkid_size[*(t+2)], 0,
            				(struct sockaddr *) &addr, addrlen);
            } else {
		u_char *raw, *s;
		int l;
		assert(*(t+2)==(find->rns?find->rns:DEFAULT_NSID));
		l=memcmp(t+4,find->rns?find->rmk:(u_char *)&find->rip,
							mkid_size[*(t+2)]);
		assert(!l);

		*t=(1<<4)|2;
		*(t+1)=0;
		*(t+3)=1;
		t=buf+4+mkid_size[*(t+2)];

		/* there might be more than one -- ignore for now */
		*t++=4; /* the only we know for now! */
		*t++=find->rns?find->rns:DEFAULT_NSID;
		s=t;
		t+=2;
		*(u_long *)t=htonl(find->begins);
		t+=sizeof(u_long);
		*(u_long *)t=htonl(find->expires);
		t+=sizeof(u_long);
		l=int_extract_raw(&raw,find->p,0); assert(l<65536);
		*t++=l>>8;
		*t++=l&0xff;
		memcpy(t,raw,l);
		t+=l;
		free(raw);
		/* from now on, alignment is not guaranteed */
		l=int_extract_raw(&raw,find->g,0); assert(l<65536);
		*t++=l>>8;
		*t++=l&0xff;
		memcpy(t,raw,l);
		t+=l;
		free(raw);
		l=int_extract_raw(&raw,find->ki,0); assert(l<65536);
		*t++=l>>8;
		*t++=l&0xff;
		memcpy(t,raw,l);
		t+=l;
		free(raw);
		l=t-s-2; assert(l<65536);
		*s++=l>>8;
		*s=l&0xff;
#if 0
                printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),*buf&15,buf[1],buf[3]);
#endif
		sendto(sock, buf, t-buf ,0 , (struct sockaddr *) &addr, 
								    addrlen);
            }
        } else {
	    *t=(1<<4)|2;
	    *(t+1)=3;
	    *(t+3)=0;
#if 0
            printf("CERT sendto(%s) %d %d %d\n",inet_ntoa(addr.sin_addr),*buf&15,buf[1],buf[3]);
#endif
            sendto(sock, buf, 4+mkid_size[*(t+2)],0, 
            				(struct sockaddr *) &addr, addrlen);
        }
    } else if ((*t & 0x0f) == 0x02 && (*(t+1) == 3 || *(t+1) == 2)) {
        /* response AND failure, just pass remote=NULL to do_discovered
         * pass it with nsid, mkid, ip to do_discovered
         */
        do_discovered(*(t+2), t+4, NULL, addr.sin_addr.s_addr);
        return;
    } 
}


/* to ask about certificates. includes own public value if available */
void udp_request(u_long addr, u_char ons, u_char *omk, u_char rns, u_char *rmk)
{
    u_char buf[20000], *t;
    int l;
    dh_data *find;
    struct sockaddr_in iaddr;
    u_char *raw;

    find=db_public_find_load(NULL,ons,omk);
    if (find == (dh_data *)(-1)) 
        find=db_public_find_load(find,ons,omk);
    t=buf;
    *t++=(1<<4)|1;
    *t++=0;
    *t++=rns;
    *t++=find?1:0;
    memcpy(t,rmk,mkid_size[rns]);
    t+=mkid_size[rns];
    if (find) {
         u_char *s;
         /* there might be more than one -- ignore for now */
         *t++=4; /* the only we know for now! */
#if USE_SUNCERT
must add x.509 certificate here (additionally)
#endif
         *t++=ons;
         s=t;
         t+=2;
         *(u_long *)t=htonl(find->begins);
	 t+=sizeof(u_long);
         *(u_long *)t=htonl(find->expires);
	 t+=sizeof(u_long);
         l=int_extract_raw(&raw,find->p,0); assert(l<65536);
         *t++=l>>8;
         *t++=l&0xff;
         memcpy(t,raw,l);
         t+=l;
         free(raw);
         /* from now on, alignment is not guaranteed */
         l=int_extract_raw(&raw,find->g,0); assert(l<65536);
         *t++=l>>8;
         *t++=l&0xff;
         memcpy(t,raw,l);
         t+=l;
         free(raw);
         l=int_extract_raw(&raw,find->ki,0); assert(l<65536);
         *t++=l>>8;
         *t++=l&0xff;
         memcpy(t,raw,l);
         t+=l;
         free(raw);
         l=t-s-2; assert(l<65536);
         *s++=l>>8;
         *s=l&0xff;
    }
    memset(&iaddr,0,sizeof (iaddr));
    iaddr.sin_family = AF_INET;
    iaddr.sin_port = htons(RECV_PORT);
    iaddr.sin_addr.s_addr = addr; /* already network format */
#if 0
    printf("CERT sendto(%s) %d %d\n",inet_ntoa(iaddr.sin_addr),*buf&15,buf[1]);
#endif
    l=sendto(send_socket,buf,t-buf,0,(struct sockaddr *)&iaddr,sizeof(iaddr));
}


